{---------------------------------------------------------------}
{  Quicksort-Metoda:                                            }
{  Nerekurzivni algoritmus s pam. slozitosti O(Nlog(N)          }
{---------------------------------------------------------------}


{ --------------------------------------------------------------}
{ Ramcovy program metody razeni                                 }
{ --------------------------------------------------------------}
PROGRAM Main;

USES CRT, DOS;

CONST N = 5000;                   { Velikost seznamu     }

TYPE  Tinfo  = char;              { Informace o polozce  }

      Titem  = record             { Vlastn prvek        }
                 key : integer;   { - klic               }
                 info: Tinfo;     { - informace o polozce}
               end;

      TListe = ARRAY [0..N] OF Titem;  { pole seznamu    }

VAR   A   : TListe;               { seznam k setrideni   }
      i,j : integer;

      h1,h2, m1,m2, s1,s2, ss1, ss2: word;
      h,m,s,ss : integer;

{---------------------------------------------------------------}
{ Partition - rozdeleni do dilcich seznamu pro Quick Sort       }
{---------------------------------------------------------------}
{ Parameter:  A    : cely seznam                                }
{             left : leva mez rozdelovane oblasti               }
{             right: prava mez rozdelovane oblasti              }
{             i    : pozice prvku Pivot                         }
{ --------------------------------------------------------------}
PROCEDURE Partition (VAR A      : TListe;
                     left, right: integer;
                     VAR i      : integer );

VAR j    : integer; { Ukazatel do seznamu k setrideni }
    pivot: TItem;   { Pivot- Prvek k rozdeleni seznamu}
    item : TItem;   { Mezipamet k provedeni zameny    }

BEGIN
  { Nastavit ukazatel behu }
  i:=left-1;
  j:=right;

  { Urcit prvek PIVOT a posunout vpravo  }
  pivot                 :=A[(left+right) SHR 1];
  A[(left+right) SHR 1] :=A[right];
  A[right]              :=pivot;

  { Utvorit dilci seznam }
  REPEAT
    REPEAT  i:=i+1 UNTIL A[i].key>=pivot.key; { i bezi  }
    REPEAT  j:=j-1 UNTIL A[j].key<=pivot.key; { j bezi  }

    IF i>=j THEN
      BEGIN { i predbehlo j, umisti polozku item   }
        item    :=A[i];
        A[i]    :=A[right];
        A[right]:=item;
      END
    ELSE
      BEGIN { zamen A[i] s A[j] }
        item:=A[i];
        A[i]:=A[j];
        A[j]:=item;
      END;

  UNTIL i>=j;   { dokud se napredhoni i a j }
END;



{ --------------------------------------------------------------}
{ Nerekurzivni algoritmus s pam. slozitosti O(Nlog(N)           }
{ --------------------------------------------------------------}

PROCEDURE QuickSort3 (VAR A: TListe; left, right: integer);

TYPE TStackItem = RECORD     { Prvek zasobniku}
                    lower: integer;
                    upper: integer;
                  END;

     Stack      = RECORD     { Zasobnik na nezpracovane polozky }
                    Top      : integer;
                    StackItem: ARRAY [1..5000] of TStackItem;
                  END;

     PStack     = ^Stack;    { Ukazatel  do zasobniku             }

VAR  S       : PStack;     { Zasobnik na nezpracovane polozky   }
     bounds  : TStackItem; { Popis polozky        [lower..upper }
     i       : Integer;    { Pivot-Pozice                     }
     pivot   : TItem;      { Prvek Pivot k rozdeleni seznamu    }
     svbound : integer;    { ulozit v mezich intervalu          }

PROCEDURE PUSH (VAR S: PStack; bounds: TStackItem);
BEGIN
  S^.Top               := S^.Top + 1;
  S^.StackItem[S^.Top] := bounds
END;


PROCEDURE POP (VAR S: PStack; VAR bounds: TStackItem);
BEGIN
  bounds := S^.StackItem[S^.Top];
  S^.Top := S^.Top - 1;
END;

Function Empty (VAR S:PStack): BOOLEAN;
BEGIN
  IF S^.Top=0 THEN Empty:=TRUE ELSE Empty:=FALSE;
END;

PROCEDURE InitStack (VAR S: PStack);
BEGIN
  New(S);
  S^.Top :=0;
END;

PROCEDURE DestroyStack (VAR S: PStack);
BEGIN
  Dispose(S)
END;

BEGIN
  { Inicializace zasobniku}
  InitStack(S);

  WITH bounds DO
    BEGIN
    { Hlavni polozku ulozit do zasobniku}
      lower := left;
      upper := right;
      Push (S,bounds);

    { Iteracni prime zpracovani vzdy vetsiho dilciho seznamu }
      WHILE NOT Empty(S) DO
        BEGIN
        { Nasledujici polozku vyjmout ze zasobniku a zpracovat ji}
          Pop(S,bounds);
          WHILE lower<upper DO
            BEGIN
              { Rozdel seznam   }
                Partition (A,lower,upper,i);

              { Vloz mensi dilci seznam do zasobniku       }
              { a hned zpracuj vetsi                       }

              IF ((i-1)-lower)<=(upper-(i+1)) THEN
                BEGIN
                 { Zapamatuj si polozku pro[lower..(i-1)] }
                   svbound := upper;
                   upper   := i-1;
                   IF lower<upper THEN Push (S,bounds);
                 { Dilci seznam [i+1..upper] ihned dal zpracovat}
                   lower:= i+1;
                   upper:= svbound;
                END
              ELSE
                BEGIN
                 { Zapamatuj si polozku pro[(i+1)..upper] }
                   svbound:=lower;
                   lower := i+1;
                   IF lower<upper THEN Push (S,bounds);
                 { Dilci seznam [lower..i-1] ihned dal zpracovat }
                   lower := svbound;
                   upper := i-1
                END
           END;
        END;
    END;
  DestroyStack(S);
END;



BEGIN {Main}

ClrScr;
Randomize;
FOR j:= 1 TO 10 DO
  BEGIN
  { Razeni}
    FOR i:=1 TO N DO A[i].key:=Random(100);

  { Razeni }
    A[0].key := -maxint;  { vlozit mezni polozku }

  { Zavolani metody razeni:                                 }
  { -----------------------                                 }
  { Quicksort3(A,1,N); : nerekurzivni metoda                }

    GetTime(h1,m1,s1,ss1);

    QuickSort3 (A,1,N);   { nerekurzivni metoda}

    GetTime(h2,m2,s2,ss2);
    ss1:= ((m2*60+s2)*100+ss2 - (m1*60+s1)*100 - ss1);
    ss := ss1 mod 100;
    s  := ss1 div 100;
    m  := s   div 60;
    h  := 0;
    WriteLn ('Cas razeni  [h M s ss]: ',h:4,m:4, s:4, ss:4);
(**********
  { Vypsat setrideny seznam }
    Write (' ':10);
    FOR i:=1 TO N DO Write (A[i].key:5); WriteLn;
***********)

  { Kontrola razeni   }
    FOR i:=1 TO N-1 DO IF (not(A[i].key<=A[i+1].key)) THEN
      BEGIN
        WriteLn ('Error detected');
        Exit
      END;
  END;
END.



