{ --------------------------------------------------------------}
{ Ramcovy program metody razeni                                 }
{ --------------------------------------------------------------}
PROGRAM Main;

USES CRT, DOS;

CONST N = 10;                   { Velikost seznamu     }

TYPE  Tinfo  = char;              { Informace o polozce  }

      Titem  = record             { Vlastn prvek        }
                 key : integer;   { - klic               }
                 info: Tinfo;     { - informace o polozce}
               end;

      TListe = ARRAY [0..N] OF Titem;  { Pole seznamu    }

VAR   A   : TListe;               { Seznam k setrideni   }
      i,j : integer;


{---------------------------------------------------------------}
{ Partition - Rozstepeni do dilcich seznamu pro Quicksort       }
{---------------------------------------------------------------}
{ Parameter:  A    : Cely seznam                                }
{             left : leva mez oblasti, kter se bude stepit     }
{             right: prava mez oblasti, kter se bude stepit    }
{             i    : pozice prvku Pivot                         }
{ --------------------------------------------------------------}

PROCEDURE Partition (VAR A      : TListe;
                     left, right: integer;
                     VAR i      : integer );

VAR j    : integer; { Ukazatel do trideneho seznamu      }
    pivot: TItem;   { Prvek Pivot k rozdeleni seznamu    }
    item : TItem;   { Mezipamet pro zamenu               }

BEGIN
  { Ukazatel behu     }
  i:=left-1;
  j:=right;

  { Urcit prvek Pivot a poposadit doprava           }
  pivot                 :=A[(left+right) SHR 1];
  A[(left+right) SHR 1] :=A[right];
  A[right]              :=pivot;

  { Vytvorit castecne seznamy}
  REPEAT
    REPEAT  i:=i+1 UNTIL A[i].key>=pivot.key; { i probiha}
    REPEAT  j:=j-1 UNTIL A[j].key<=pivot.key; { j probiha}

    IF i>=j THEN
      BEGIN { i predbehlo j , nastavena polozka Pivot   }
        item    :=A[i];
        A[i]    :=A[right];
        A[right]:=item;
      END
    ELSE
      BEGIN { zamen A[i]  s  A[j] }
        item:=A[i];
        A[i]:=A[j];
        A[j]:=item;
      END;

  UNTIL i>=j;   { dokud se i a j nepredhoni      }
END;


{ --------------------------------------------------------------}
{ Rekurzivni metoda QuickSort                                   }
{ --------------------------------------------------------------}

PROCEDURE QuickSort1 (VAR A: TListe; left, right: integer);

VAR i : integer; { Ukazatel na pozici Pivot  }

BEGIN
  IF (left<right) THEN { k dispozici nejmene dve polozky }
    BEGIN
      { Rozdel seznam   }
        Partition (A,left,right,i);

      { Roztrid oba zbyvajici seznamy  }
        Quicksort1 (A,left ,i-1);
        Quicksort1 (A,i+1,right);
    END
END;



{ --------------------------------------------------------------}
{ Quicksort - Metoda   s pametovou slozitosti O(log(N))         }
{ --------------------------------------------------------------}

PROCEDURE QuickSort2 (VAR A: TListe; left, right: integer);

VAR i : integer; { Ukazatel na pozici Pivot  }

BEGIN
{ Iteracni prime zpracovani uz vetsich                                 }
{ castecnych seznamu v pripade, ze mame nejmene dve polozky k dispozici}
  WHILE (left<right) DO
    BEGIN
      { Rozdel seznam   }
        Partition (A,left,right,i);

      { Rekurzivne volej Quicksort pro mensi dilci seznam      }
      { a hned zpracuj vetsi                                   }

      IF ((i-1)-left)<=(right-(i+1)) THEN
        BEGIN
          Quicksort2(A,left,i-1);
          left := i+1
        END
      ELSE
        BEGIN
          Quicksort2(A,i+1,right);
          right := i-1;
        END;
    END
END;


{ --------------------------------------------------------------}
{ Nerekurzivni metoda Quicksort s pametovou slozitosti O(Nlog(N)}
{ --------------------------------------------------------------}

PROCEDURE QuickSort3 (VAR A: TListe; left, right: integer);

TYPE TStackItem = RECORD     { Prvek zasobniku}
                    lower: integer;
                    upper: integer;
                  END;

     Stack      = RECORD     { Zasobnik na nazpracovane polozky }
                    Top      : integer;
                    StackItem: ARRAY [1..N] of TStackItem;
                  END;

     PStack     = ^Stack;    { Ukazatel do zasobniku           }

VAR  S       : PStack;     { Zasobnik na nazpracovane polozky   }
     bounds  : TStackItem; { Popis polozky [lower..upper }
     i       : Integer;    { Pozice Pivot                       }
     pivot   : TItem;      { Prvek Pivot k rozdeleni seznamu    }
     svbound : integer;    { hlida docasne meze intervalu       }

PROCEDURE PUSH (VAR S: PStack; bounds: TStackItem);
BEGIN
  S^.Top               := S^.Top + 1;
  S^.StackItem[S^.Top] := bounds
END;

PROCEDURE POP (VAR S: PStack; VAR bounds: TStackItem);
BEGIN
  bounds := S^.StackItem[S^.Top];
  S^.Top := S^.Top - 1;
END;

Function Empty (VAR S:PStack): BOOLEAN;
BEGIN
  IF S^.Top=0 THEN Empty:=TRUE ELSE Empty:=FALSE;
END;

PROCEDURE InitStack (VAR S: PStack);
BEGIN
  New(S);
  S^.Top :=0;
END;

PROCEDURE DestroyStack (VAR S: PStack);
BEGIN
  Dispose(S)
END;


BEGIN
  { Inicializace zasobniku}
  InitStack(S);

  WITH bounds DO
    BEGIN
    { Vlozit hlavni polozku do zasobniku}
      lower := left;
      upper := right;
      Push (S,bounds);

    { Iteracni prime zpracovani uz vetsich dilcich seznamu  }
      WHILE NOT Empty(S) DO
        BEGIN
        { Vzit dal polozku ze sasobniku a zpracovat  }
          Pop(S,bounds);
          WHILE lower<upper DO
            BEGIN
              { Rozdel seznam   }
                Partition (A,lower,upper,i);

              { Vloz mensi dilci seznam do zasobniku       }
              { a hned zpracuj vetsi                       }

              IF ((i-1)-lower)<=(upper-(i+1)) THEN
                BEGIN
                 { zapamatovat si polozku pro [lower..(i-1)]  }
                   svbound := upper;
                   upper   := i-1;
                   IF lower<upper THEN Push (S,bounds);
                 { dilci seznam [i+1..upper] ihned dal zpracovat}
                   lower:= i+1;
                   upper:= svbound;
                END
              ELSE
                BEGIN
                 { Zapamatovat si [(i+1)..upper]  }
                   svbound:=lower;
                   lower := i+1;
                   IF lower<upper THEN Push (S,bounds);
                 { Dilci seznam [lower..i-1] ihned dal zpracovat}
                   lower := svbound;
                   upper := i-1
                END
           END;
        END;
    END;
  DestroyStack(S);
END;

{ --------------------------------------------------------------}
{ Rekurzivni metoda Quicksort     (Wirthova varianta)           }
{ --------------------------------------------------------------}

PROCEDURE QuickSort4 (VAR A: TListe; left,right: integer);

VAR i,j: integer;  item, pivot: TItem;

BEGIN
  IF (left<right) THEN
    BEGIN
      i:=left;
      j:=right;
      pivot := A[(left+right) DIV 2];

      REPEAT
        WHILE A[i].key<pivot.key DO i:=i+1;
        WHILE A[j].key>pivot.key DO j:=j-1;

        IF i<=j THEN BEGIN
                       item :=A[i];
                       A[i] :=A[j];
                       A[j] :=item;
                       i    :=i+1;
                       j    :=j-1;
                     END
        UNTIL i>j;

        IF left<j  THEN Quicksort4 (A,left ,j);
        IF i<right THEN Quicksort4 (A,i,right);
    END
END;


BEGIN {Main}

ClrScr;
Randomize;
FOR j:= 1 TO 100 DO
  BEGIN
  { Predobsadit seznam   }
    FOR i:=1 TO N DO A[i].key:=Random(100);

  { Razeni    }
    A[0].key := -maxint;  { Stopper-Item eintragen }

  { Volani metody razeni     :                                 }
  { --------------------------                                 }
  { Quicksort1(A,1,N); : rekurzivni metoda                     }
  { Quicksort2(A,1,N); : rekurzivni s pam. slozitosti O(log(N))}
  { Quicksort3(A,1,N); : nerekurzivni metoda                   }
  { Quicksort4(A,1,N); : rekurzivni metoda podle Wirtha        }

    QuickSort1 (A,1,N);

  { Kontrola razeni   }
    FOR i:=1 TO N-1 DO IF (not(A[i].key<=A[i+1].key))
      THEN BEGIN
             WriteLn ('Error detected');
             Exit
           END;
    WriteLn ('Seznam ', j, '  setriden !');
  END;
END.




